@@ -5,7 +5,8 @@ import glob
 import logging
 import os
 import time
-from concurrent.futures import ThreadPoolExecutor
+from concurrent.futures import ThreadPoolExecutor, Future
+from functools import partial
 from pathlib import Path
 from typing import Optional, Generator, Iterable, List
 
@@ -14,6 +15,7 @@ import botocore
 import yaml
 from botocore.exceptions import ClientError, ValidationError, WaiterError
 from deepmerge import always_merger
+from tqdm import tqdm
 from yaml import Loader
 from cf_deploy.utils.logging import configure_structlog
 import structlog
@@ -38,7 +40,7 @@ def generic_constructor(loader, tag_suffix, node):
 yaml.SafeLoader.add_multi_constructor(u'!', generic_constructor)
 
 
-def track_stack_events(stack_name, region):
+def track_stack_events(stack_name, region, verbose=True):
     cf = boto3.client('cloudformation', region_name=region)
     start = time.time()
     seen_event_ids = set()
@@ -49,7 +51,7 @@ def track_stack_events(stack_name, region):
         for event in reversed(events['StackEvents']):
             event_id = event['EventId']
             if event_id not in seen_event_ids and event['Timestamp'].timestamp() > start:
-                log.info(
+                log.msg(logging.INFO if verbose else logging.DEBUG,
                     "Stack event",
                     timestamp=event['Timestamp'],
                     resource_status=event['ResourceStatus'],
@@ -67,7 +69,7 @@ def track_stack_events(stack_name, region):
         time.sleep(5)
 
 
-def create_change_stack(stack_name, config: Config, base_config: BaseConfig) -> Optional[str]:
+def create_change_stack(stack_name, config: Config, base_config: BaseConfig, verbose=True) -> Optional[str]:
     cf = boto3.client('cloudformation', region_name=config.region)
 
     # Describe stacks with name
@@ -83,16 +85,16 @@ def create_change_stack(stack_name, config: Config, base_config: BaseConfig) ->
     ), None)
 
     if stack and "IN_PROGRESS" in stack["StackStatus"] and "REVIEW_IN_PROGRESS" not in stack["StackStatus"]:
-        log.warning("Stack is already in progress", name=stack_name)
+        (log.info if verbose else log.debug)("Stack is already in progress", name=stack_name)
         track_stack_events(stack_name, config.region)
 
     # Parse template body
     template_yaml = yaml.safe_load(config.template)
     # parameter_keys = [p["ParameterKey"] for p in template_yaml.get("Parameters", {}).keys()]
-    parameter_keys = template_yaml["Parameters"].keys()
+    parameter_keys = template_yaml.get("Parameters", {}).keys()
 
     # Create change set
-    log.info("Creating change set", name=stack_name)
+    (log.info if verbose else log.debug)("Creating change set", name=stack_name)
     try:
         change_set_response = cf.create_change_set(
             StackName=stack_name,
@@ -117,7 +119,7 @@ def create_change_stack(stack_name, config: Config, base_config: BaseConfig) ->
     except ClientError as e:
         error = e.response["Error"]
         if error["Code"] == "ValidationError" and "No updates are to be performed" in error["Message"]:
-            log.info("No changes to deploy", name=stack_name)
+            (log.info if verbose else log.debug)("No changes to deploy", name=stack_name)
             return
         log.error(f"Failed to create change set: {error['Message']}", name=stack_name)
         return
@@ -139,7 +141,7 @@ def create_change_stack(stack_name, config: Config, base_config: BaseConfig) ->
 
     if change_set['Status'] == 'FAILED':
         if "The submitted information didn't contain changes" in change_set['StatusReason']:
-            log.info("No changes to deploy", name=stack_name)
+            (log.info if verbose else log.debug)("No changes to deploy", name=stack_name)
             return
 
         log.error("Change set failed", reason=change_set['StatusReason'])
@@ -147,7 +149,7 @@ def create_change_stack(stack_name, config: Config, base_config: BaseConfig) ->
 
     # Print changes
     for change in change_set['Changes']:
-        log.info(
+        log.msg(logging.INFO if verbose else logging.DEBUG,
             "Change",
             resource_type=change['ResourceChange']['ResourceType'],
             action=change['ResourceChange']['Action'],
@@ -158,11 +160,11 @@ def create_change_stack(stack_name, config: Config, base_config: BaseConfig) ->
     return change_set_response['Id']
 
 
-def deploy_stack(stack_name, config: Config, base_config: BaseConfig, arguments):
+def deploy_stack(stack_name, config: Config, base_config: BaseConfig, arguments, verbose=True):
     try:
-        change_set_id = create_change_stack(stack_name, config, base_config)
+        change_set_id = create_change_stack(stack_name, config, base_config , verbose=verbose)
     except Exception as e:
-        log.exception(e)
+        log.exception(e, stack_name=stack_name)
         return
 
     if arguments.dry_run or not change_set_id:
@@ -173,7 +175,8 @@ def deploy_stack(stack_name, config: Config, base_config: BaseConfig, arguments)
             log.info("Aborting deployment")
             return
 
-    log.info("Deploying", name=stack_name)
+    (log.info if verbose else log.debug)("Deploying", name=stack_name)
+
     cf = boto3.client('cloudformation', region_name=config.region)
     cf.execute_change_set(
         StackName=stack_name,
@@ -181,7 +184,9 @@ def deploy_stack(stack_name, config: Config, base_config: BaseConfig, arguments)
     )
 
     if not arguments.skip_wait:
-        track_stack_events(stack_name, config.region)
+        track_stack_events(stack_name, config.region, verbose=verbose)
+
+    (log.info if verbose else log.debug)("Finished Deployment", name=stack_name)
 
 
 def loading_config(path: Path, base_config: BaseConfig, arguments) -> Iterable[Config]:
@@ -257,6 +262,7 @@ def main():
     parser.add_argument("--skip-wait", help="Disable waiting for stack to be deployed", action="store_true")
     parser.add_argument("--parallel", help="Deploy stacks in parallel", action="store_true")
     parser.add_argument("--delete-deprecated", help="Delete stacks that are not in the config", action="store_true")
+    parser.add_argument("--concurrency", help="Number of stacks to deploy in parallel", default=16, type=int)
 
     args = parser.parse_args()
 
@@ -275,10 +281,21 @@ def main():
     # Load configs
     if args.parallel:
         log.info("Deploying stacks in parallel")
-        with ThreadPoolExecutor(max_workers=10) as executor:
+        with ThreadPoolExecutor(max_workers=args.concurrency) as executor, tqdm(total=len(configs), desc="Stacks", unit="stack") as progress_bar:
+            futures: List[Future] = []
             for config in configs:
                 stack_name = f"{base_config.prefix or ''}{config.name}"
-                executor.submit(deploy_stack, stack_name, config, base_config, args)
+                futures.append(executor.submit(deploy_stack, stack_name, config, base_config, args, False))
+
+            while futures:
+                for future in list(futures):
+                    if future.done():
+                        progress_bar.update(1)
+                        futures.remove(future)
+
+                progress_bar.display()
+                time.sleep(0.5)
+
     else:
         for config in configs:
             stack_name = f"{base_config.prefix or ''}{config.name}"
